'******************************************************************************
' Module:  OSHandle.vb
' Notices: Copyright (c) 2002 Jeffrey Richter
'******************************************************************************


Imports System


'//////////////////////////////////////////////////////////////////////////////


' Implementing the IDisposable interface signals users of
' this class that it offers the dispose pattern.
Public Class OSHandle
   Implements IDisposable

   ' This field holds the Win32 handle of the unmanaged resource.
   Private handle As IntPtr

   ' This constructor initializes the handle.
   Public Sub New(ByVal handle As IntPtr)
      Me.handle = handle
   End Sub

   ' When garbage collected, this Finalize method, which
   ' will close the unmanaged resources handle, is called.
   Protected Overrides Sub Finalize()
      Dispose(False)
   End Sub

   ' This public method can be called to deterministically
   ' close the unmanaged resources handle.
   Public Sub Dispose() Implements IDisposable.Dispose
      ' Because the object is explicitly cleaned up, stop the
      ' garbage collector from calling the Finalize method.
      GC.SuppressFinalize(Me)

      ' Call the method that actually does the cleanup.
      Dispose(True)
   End Sub

   ' This public method can be called instead of Dispose.
   Public Sub Close()
      Dispose()
   End Sub

   ' The common method that does the actual cleanup.
   ' Finalize, Dispose, and Close call this method.
   ' Because this class isnt NotInheritable, this method is protected.
   Protected Overridable Sub Dispose(ByVal disposing As Boolean)
      ' Synchronize threads calling Dispose/Close simultaneously.
      SyncLock Me
         If disposing Then
            ' The object is being explicitly disposed of/closed, not
            ' finalized. It is therefore safe for code in this If
            ' statement to access fields that reference other
            ' objects because the Finalize method of these other objects
            ' hasnt yet been called.
            ' For the OSHandle class, there is nothing to do in here.
         End If
         ' The object is being disposed of/closed or finalized.
         If IsValid Then
            ' If the handle is valid, close the unmanaged resource.
            ' NOTE: Replace CloseHandle with whatever function is
            ' necessary to close/free your unmanaged resource.
            CloseHandle(handle)
            ' Set the handle field to some sentinel value. This precaution
            ' prevents the possibility of calling CloseHandle twice.
            handle = InvalidHandle
         End If
      End SyncLock
   End Sub

   ' Public property to return the value of an invalid handle.
   ' NOTE: Make this property return an invalid value for
   ' whatever unmanaged resource youre using.
   Public ReadOnly Property InvalidHandle() As IntPtr
      Get
         Return IntPtr.Zero
      End Get
   End Property

   ' Public method to return the value of the wrapped handle
   Public Function ToHandle() As IntPtr
      Return handle
   End Function

   ' Public properties to return whether the wrapped handle is valid
   Public ReadOnly Property IsValid() As Boolean
      Get
         Return Not handle.Equals(InvalidHandle)
      End Get
   End Property

   Public ReadOnly Property IsInvalid() As Boolean
      Get
         Return Not IsValid
      End Get
   End Property

   ' Private method called to free the unmanaged resource
   <System.Runtime.InteropServices.DllImport(Kernel32)> _
   Protected Shared Function CloseHandle(ByVal handle As IntPtr) As Boolean
   End Function
End Class


'/////////////////////////////// End of File //////////////////////////////////
