'******************************************************************************
' Module:  SimpleBinder.vb
' Notices: Copyright (c) 2002 Jeffrey Richter
' Thanks:  To Dario Russi for supplying the initial version of this code.
'******************************************************************************


Imports System
Imports System.Reflection
Imports System.Collections
Imports CultureInfo = System.Globalization.CultureInfo
    

'//////////////////////////////////////////////////////////////////////////////


Public NotInheritable Class SimpleBinder 
   Inherits Binder

   ' Type.InvokeMember calls this method if more than 1 field matches.
   ' This code performs uses simple conversion rules to bind.
   Public Overrides Function BindToField( _
      ByVal bindingAttr As BindingFlags, _
      ByVal fields() As FieldInfo, _
      ByVal value As Object, _
      ByVal culture As CultureInfo) _
      As FieldInfo

      ' bindingAttr: Flags to restrict options
      ' fields(): Field subset selected by reflection
      ' value: The value to set 
      ' culture: Culture (usually ignored)

      ' Get the type of the value that is to be assigned to the field
      Dim valueType As Type = value.GetType()

      ' If any fields exactly match the value's type, return that field
      Dim f as FieldInfo
      For Each f In fields
         If (valueType Is f.FieldType) Then Return f
      Next

      ' If any fields have a "compatible" type, return that field
      For Each f In fields
         ' Get the type of the field
         Dim formalType As Type = f.FieldType

         ' If the value's type is compatible with the field's type, return it
         If CanConvertPrimitiveType(valueType, formalType) Then Return f

         ' Consider a match if the value can be assigned to the field's type
         If Not formalType.IsValueType Then
            ' It must be some sort of "compatible" reference type
            If (formalType.IsAssignableFrom(valueType)) Then Return f
         End If
      Next

      ' No compatible field was found
      Throw New MissingFieldException("Field not found.")
   End Function


   ' Type.InvokeMember and Activator.CreateInstance call this method to
   ' select a specific method. 
   ' This code performs uses simple conversion rules to bind.
   Public Overrides Function BindToMethod( _
      ByVal bindingAttr As BindingFlags, _
      ByVal methods() As MethodBase, _
      ByRef args() As Object, _
      ByVal modifiers() As ParameterModifier, _
      ByVal culture As CultureInfo, _
      ByVal names() As String, _
      ByRef state As Object) _
      As MethodBase

      ' bindingAttr: Flags to restrict options
      ' methods: Method subset selected by reflection
      ' args: Arguments provided by the caller (BindToMethod can modify this array)
      ' modifiers: Modifiers (usually ignored)
      ' culture: Culture (usually ignored)
      ' names: Named arguments (if any)
      ' state: If 'args' is changed, this object can be used to get 
      ' back the original array by calling ReorderArgumentArray

      ' This binder doesn't support argument re-ordering
      state = Nothing

      ' Construct array of method argument types.
      Dim argType(args.Length - 1) As Type
      Dim i As Int32
      For i = 0 To args.Length - 1
         If Not args(i) Is Nothing Then
            argType(i) = args(i).GetType()
         End If
      Next

      ' A more sophisticated binder would have code here to deal with methods
      ' that accept a variable number of arguments (ParamArrayAttribute) and 
      ' with methods that accept optional arguments and named parameters.

      ' Select a method that matches type argument's types.
      Return SelectMethod(bindingAttr, methods, argType, modifiers)
   End Function


   ' Flags indicating how to compare the specified argument types 
   ' with the method's parameter types.
   <Flags()> _
   Private Enum CompareParamAndArgTypesFlags
      Exact = &H0
      CoerceValueTypes = &H1
      AllowBaseTypes = &H2
   End Enum


   ' This method returns true if the specified argument types match
   ' the method's parameter types
   Private Shared Function CompareParamAndArgTypes( _
      ByVal paramTypes() As ParameterInfo, _
      ByVal argTypes() As Type, _
      ByVal flags As CompareParamAndArgTypesFlags) _
      As Boolean

      ' This binder requires that the number of arguments and parameters match
      If (paramTypes.Length <> argTypes.Length) Then Return False

      Dim i As Int32
      For i = 0 To paramTypes.Length - 1
         ' If the argument has a type, compare it against the parameter's type
         ' This can be null if Type.InvokeMember is passed null for an argument
         If Not argTypes(i) Is Nothing Then
            Dim formalType As Type = paramTypes(i).ParameterType

            ' If argument and parameter types match exactly, try next pair
            If formalType Is argTypes(i) Then GoTo LoopAround

            ' Compare the primitive, value type, or enumerated type parameter
            If formalType.IsValueType Then

               If ((flags And CompareParamAndArgTypesFlags.CoerceValueTypes) <> 0) And _
                   CanConvertPrimitiveType(argTypes(i), formalType) Then GoTo LoopAround
               Exit For ' Can't coerce argument type to parameter type
            Else

               ' Compare the reference type parameter
               If ((flags And CompareParamAndArgTypesFlags.AllowBaseTypes) <> 0) And _
                   formalType.IsAssignableFrom(argTypes(i)) Then GoTo LoopAround
               Exit For ' Can't implicitly cast argument type to parameter type
            End If
         End If
LoopAround:
      Next

      ' Return true if all argument and parameter types match
      Return (i = paramTypes.Length)
   End Function


   ' Called to change a type during invocation.
   ' There must have been a type mismatch and we are asked to intervene.
   Public Overrides Function ChangeType(ByVal value As Object, ByVal t As Type, ByVal culture As CultureInfo) As Object
      ' We only do primitive conversions.
      If CanConvertPrimitiveType(value.GetType(), t) Then
         Return DoConvertPrimitiveType(value, t)
      End If
      Throw New ArgumentException("No conversion allowed for one of the arguments")
   End Function


   ' Called to restore the args array back to that passed to BindToMethod.
   ' This code does nothing because we don't handle named or optional arguments
   Public Overrides Sub ReorderArgumentArray( _
      ByRef args() As Object, _
      ByVal state As Object)
      ' args: Arguments provided by the caller 
      ' state: This object can is used to get back the original array

      ' Here's the scenario where this method comes in useful...
      ' Say InvokeMember is called passing some named arguments. The elements
      ' in this array must be reordered so that arguments are in the correct 
      ' position before invoking the method. For optional arguments, the
      ' array must grow to accommodate the unspecified arguments.

      ' Note: If an argument is marked as 'out' or 'ref', the return value 
      ' will be updated in this array. These array element values must be 
      ' copied back to the reflection caller's original array so that they 
      ' can get the 'returned' values.

      ' For example, let's say that Class Foo defines the following method:
      ' Class Foo
      '    Public Shared Sub m(i as Int32, ByRef o As Object)
      '       ...
      '    End Sub
      ' End Class

      ' Now, let's invoke this method using a named parameter as follows:
      ' Object[] args = new Object[] { obj, 3 };
      ' typeof(Foo).InvokeMember("m", 
      '    BindingFlags.InvokeMethod Or BindingFlags.Static Or BindingFlags.Public,
      '    New SimpleBinder(),  // The binder
      '    Nothing,                // No object; static method
      '    args,                // The arguments to pass (Object followed by Int32)
      '    Nothing,                // No ParamterModifier array
      '    Nothing,                // No Culture
      '    New String() {"o"}); // The first argument should be the 'o' parameter

      ' Notice the Object and Int32 arguments are passed in reverse order. 
      ' This is OK if the binder can deal with named parameters.

      ' In BindToMethod, the argument array must be reordered in order to 
      ' sucessfully invoke the method. This array can't be returned to the 
      ' caller since the caller would have no knowledge what was where. 
      ' Also, the caller needs to fetch the 'ref' value out of the array.

      ' This ReorderArgumentArray method must transform the array used to 
      ' invoke the method back to the array expected by the caller.

      ' BindToMethod should save any information required to restore the 
      ' argument array in the state parameter passed to this method.
   End Sub


   ' GetMethod calls this method to select a specific method. 
   ' This code performs uses simple conversion rules to bind.
   Public Overrides Function SelectMethod( _
      ByVal bindingAttr As BindingFlags, _
      ByVal methods() As MethodBase, _
      ByVal argTypes() As Type, _
      ByVal modifiers() As ParameterModifier) _
      As MethodBase

      ' bindingAttr: Flags to restrict options
      ' methods: Method subset selected by reflection
      ' argTypes: Set of argument types
      ' modifiers: Modifiers (usually ignored)

      ' This ArrayList contains the set of possible methods
      Dim candidates As ArrayList = New ArrayList()

      ' Build the set of candidate methods removing any method that
      ' doesn't have the specified number of arguments.

      ' A more sophisticated binder would have code here to deal with methods
      ' that accept a variable number of arguments (ParamArrayAttribute) and 
      ' with methods that accept optional arguments and named parameters.

      Dim argCount As Int32 = argTypes.Length
      Dim m As MethodBase
      For Each m In methods
         If m.GetParameters().Length = argCount Then candidates.Add(m)
      Next

      ' If no exact match was found and the caller wants an exact match, throw
      Dim numTests As Int32 = 1
      If (BindingFlags.ExactBinding And bindingAttr) = 0 Then numTests = 3

      Dim test As Int32
      For test = 0 To numTests - 1
         Dim flags As CompareParamAndArgTypesFlags = CompareParamAndArgTypesFlags.Exact

         ' Look for methods whose parameter types exactly match the argument types
         If test = 0 Then flags = CompareParamAndArgTypesFlags.Exact

         ' Look for methods that can accomodate the parameters passed.
         If test = 1 Then flags = CompareParamAndArgTypesFlags.AllowBaseTypes

         ' Check whether any conversion could be applied on primitives as well. 
         ' NOTE: We bind to the first matching method; not the best matching method.
         ' Look for methods that can accomodate the parameters passed.
         If test = 2 Then
            flags = CompareParamAndArgTypesFlags.AllowBaseTypes Or _
               CompareParamAndArgTypesFlags.CoerceValueTypes
         End If

         ' Assume that no method matches the argument's parameters
         Dim match As MethodBase = Nothing

         ' Check ALL of the methods
         For Each m In candidates

            If CompareParamAndArgTypes(m.GetParameters(), argTypes, flags) Then
               ' A matching method was found

               ' If we found a matching method previously, then we don't know
               ' how to select one of them.
               If Not match Is Nothing Then
                  ' Note, when doing an exact match this can occur if multiple 
                  ' methods differ only by return type
                  Throw New AmbiguousMatchException("Multiple methods match the specified parameter types.")
               End If

               ' Save the matching method
               match = m
            End If
         Next

         ' One matching method was found, return it
         If Not match Is Nothing Then Return match
      Next

      ' No matching method was found, throw
      Throw New MissingMethodException("Member not found.")
   End Function


   ' Called to return the parameters of a property
   Private Shared Function GetPropertyParams(ByVal p As PropertyInfo) As ParameterInfo()

      ' If a get accessor method exists, return its parameters
      Dim m As MethodInfo = p.GetGetMethod()
      If Not m Is Nothing Then Return m.GetParameters()

      ' No get accessor method exists, use the set access method
      m = p.GetSetMethod()
      Dim setParams() As ParameterInfo = m.GetParameters()

      ' Copy all elements but the last (the property's type) to a new array
      Dim paramTypes(setParams.Length - 2) As ParameterInfo ' = New ParameterInfo()
      Array.Copy(setParams, paramTypes, paramTypes.Length)

      ' Return the copy
      Return paramTypes
   End Function


   ' GetProperty calls this method to select a specific property. 
   ' This code performs uses simple conversion rules to bind.
   Public Overrides Function SelectProperty( _
      ByVal bindingAttr As BindingFlags, _
      ByVal properties() As PropertyInfo, _
      ByVal returnType As Type, _
      ByVal argTypes() As Type, _
      ByVal modifiers() As ParameterModifier) _
      As PropertyInfo

      ' bindingAttr: Flags to restrict options
      ' properties: Property subset selected by reflection
      ' returnType: Property's return type
      ' argTypes: Set of argument types
      ' modifiers: Modifiers (usually ignored)

      ' This ArrayList contains the set of possible properties
      Dim candidates As ArrayList = New ArrayList()

      ' Build the set of candidate properties removing any property that 
      ' doesn't have the specified number of arguments.

      ' Only consider properties that have the same number of arguments and type
      Dim argCount As Int32
      If argTypes Is Nothing Then argCount = 0 Else argCount = argTypes.Length
      Dim p As PropertyInfo
      For Each p In properties
         If GetPropertyParams(p).Length = argCount Then
            ' Check the property's type
            If (returnType Is Nothing) Or (returnType Is p.PropertyType) Then
               candidates.Add(p)
            End If
         End If
      Next

      ' If no exact match was found and the caller wants an exact match, throw
      Dim numTests As Int32
      If (BindingFlags.ExactBinding And bindingAttr) <> 0 Then numTests = 1 Else numTests = 3

      Dim test As Int32
      For test = 0 To numTests - 1
         Dim flags As CompareParamAndArgTypesFlags = CompareParamAndArgTypesFlags.Exact

         ' Look for properties whose parameter types exactly match the argument types
         If test = 0 Then flags = CompareParamAndArgTypesFlags.Exact

         ' Look for properties that can accomodate the parameters passed.
         If test = 1 Then flags = CompareParamAndArgTypesFlags.AllowBaseTypes

         ' Check whether any conversion could be applied on primitives as well. 
         ' NOTE: We bind to the first matching property; not the best matching property.
         ' Look for properties that can accomodate the parameters passed.
         If test = 2 Then flags = CompareParamAndArgTypesFlags.AllowBaseTypes Or _
            CompareParamAndArgTypesFlags.CoerceValueTypes

         ' Assume that no property matches the argument's parameters
         Dim match As PropertyInfo = Nothing

         ' Check ALL of the properties
         For Each p In candidates

            If CompareParamAndArgTypes(GetPropertyParams(p), argTypes, flags) Then
               ' A matching property was found

               ' If we found a matching property previously, then we don't know
               ' how to select one of them.
               If Not match Is Nothing Then
                  ' Note, when doing an exact match this can occur if multiple 
                  ' properties differ only by type
                  Throw New AmbiguousMatchException("Multiple properties match the specified parameter types.")
               End If

               ' Save the matching property
               match = p
            End If
         Next

         ' One matching property was found, return it
         If Not match Is Nothing Then Return match
      Next

      ' No matching property was found, throw
      Throw New MissingMemberException("Member not found.")
   End Function


   '//////////////////////////////////////////////////////////////////////////////


   ' This table indicates what conversions this binder allows
   Shared ReadOnly AllowedConversions()() As TypeCode = New TypeCode(18)() {}

   ' This static constructor initializes the conversion table
   Shared Sub New()
      ' For example, Char can be convert to SByte, Byte, or UInt16
      AllowedConversions(CType(TypeCode.Char, Int32)) = New TypeCode() _
         {TypeCode.SByte, TypeCode.Byte, TypeCode.UInt16}

      AllowedConversions(CType(TypeCode.Int16, Int32)) = New TypeCode() _
         {TypeCode.SByte, TypeCode.Byte}

      AllowedConversions(CType(TypeCode.UInt16, Int32)) = New TypeCode() _
         {TypeCode.Char, TypeCode.SByte, TypeCode.Byte}

      AllowedConversions(CType(TypeCode.Int32, Int32)) = New TypeCode() _
         {TypeCode.Char, TypeCode.SByte, TypeCode.Byte, TypeCode.Int16, TypeCode.UInt16}

      AllowedConversions(CType(TypeCode.UInt32, Int32)) = New TypeCode() _
         {TypeCode.Char, TypeCode.SByte, TypeCode.Byte, TypeCode.Int16, TypeCode.UInt16}

      AllowedConversions(CType(TypeCode.Int64, Int32)) = New TypeCode() _
         {TypeCode.Char, TypeCode.SByte, TypeCode.Byte, TypeCode.Int16, TypeCode.UInt16, TypeCode.Int32, TypeCode.UInt32}

      AllowedConversions(CType(TypeCode.UInt64, Int32)) = New TypeCode() _
         {TypeCode.Char, TypeCode.SByte, TypeCode.Byte, TypeCode.Int16, TypeCode.UInt16, TypeCode.Int32, TypeCode.UInt32}

      AllowedConversions(CType(TypeCode.Single, Int32)) = New TypeCode() _
         {TypeCode.Char, TypeCode.SByte, TypeCode.Byte, TypeCode.Int16, TypeCode.UInt16}

      AllowedConversions(CType(TypeCode.Double, Int32)) = New TypeCode() _
         {TypeCode.Char, TypeCode.SByte, TypeCode.Byte, TypeCode.Int16, TypeCode.UInt16, TypeCode.Int32, TypeCode.UInt32, TypeCode.Single}
   End Sub


   ' Returns 'true' if the binder can convert from fromType to toType
   Private Shared Function CanConvertPrimitiveType(ByVal fromType As Type, ByVal toType As Type) As Boolean
      ' If the types are the same, of course we can convert
      If fromType Is toType Then Return True

      ' Check table to see if fromType can be converted to anything
      Dim fromTypeCode As TypeCode = Type.GetTypeCode(fromType)
      If AllowedConversions(CType(fromTypeCode, Int32)) Is Nothing Then Return False

      ' Check table to see if conversion from fromType to toType is allowed
      Dim toTypeCode As TypeCode = Type.GetTypeCode(toType)
      If Array.IndexOf(AllowedConversions(CType(toTypeCode, Int32)), fromTypeCode) <> -1 Then
         Return True
      End If

      ' Conversion is not allowed
      Return False
   End Function


   ' Returns new object converted from original type
   Private Shared Function DoConvertPrimitiveType(ByVal value As Object, ByVal toType As Type) As Object
      ' If conversion isn't allowed, throw
      If Not CanConvertPrimitiveType(value.GetType(), toType) Then
         Throw New InvalidCastException()
      End If

      ' Conversion is allowed, convert and return the new object
      Return Convert.ChangeType(value, toType)
   End Function
end class


'/////////////////////////////// End of File //////////////////////////////////
